import numpy as np

PI = np.pi
"""Pi constant."""

FREE_SPACE_PERMITTIVITY: float = 8.85418782e-12
"""Permittivity of free space [m-3 kg-1 s4 A2]."""

FREE_SPACE_PERMEABILITY: float = 4 * PI * 1e-7
"""Permeability of free space [m kg s-2 A-2]."""

FREE_SPACE_LIGHT_SPEED: float = 299792458.0
"""Speed of light in free space [m/s]."""
